package org.test;

import java.util.*;
import javax.persistence.*;


public final class Main {

	public static void main(String[] args) {
		EntityManagerFactory emf = Persistence.createEntityManagerFactory("objectdb:$objectdb/db/test.tmp;drop");
		EntityManager em = emf.createEntityManager();

		em.getTransaction().begin();
		String mainType = "MAIN";
		String subType = "SUB";

		for(int i = 0; i < 2000; i++) {
			Element element = new Element(mainType, null);
			element.addCostObject(new CostObject(2));
			element.addCostObject(new CostObject(3));
			element.addCostObject(new CostObject(4));
			em.persist(element);

			for(int j = 0; j < 5; j++) {
				Element childElement = new Element(subType, element);
				childElement.addCostObject(new CostObject(2));
				childElement.addCostObject(new CostObject(3));
				childElement.addCostObject(new CostObject(4));
				em.persist(childElement);
			}
		}
		em.getTransaction().commit();
		em.close();

		em = emf.createEntityManager();
		List<Element> allElements = em.createQuery("SELECT e FROM Element e where e.type = :type").setParameter("type", "MAIN").getResultList();

		ArrayList<CostObject> costObjects = new ArrayList<>();
		for(Element element : allElements) {
			costObjects.addAll(element.getAllCostObjects());
		}

		System.out.println(costObjects.size() + " should be: " + 36000);

		try {
			long totalAmount = 0;
			for(Element element : allElements) {
				totalAmount += element.getAmountTotal();
			}

			System.out.println(totalAmount + " should be: " + 108000);
		} catch(NullPointerException | ConcurrentModificationException exception) {
			System.out.println("Should not throw exception");
		}

		em.close();
		emf.close();
	}

	@Entity
	public static class MyEntity {
		private String name;

		MyEntity(String name) {
			this.name = name;
		}

		@Override
		public String toString() {
			return name;
		}
	}

	@Embeddable
	public static class CostObject {
		@Basic
		private long amount;

		public CostObject(long amount) {
			this.amount = amount;
		}

		public long getAmount() {
			return amount;
		}
	}

	@MappedSuperclass
	@SequenceGenerator(name = "seq", initialValue = 1, allocationSize = 100)
	public static abstract class BaseElement {

		@Id
		@GeneratedValue(strategy = GenerationType.SEQUENCE, generator = "seq")
		private Long id;
	}

	@MappedSuperclass
	public static abstract class BaseCostElement extends BaseElement {

		@ElementCollection(fetch = FetchType.EAGER)
		private List<CostObject> costObjects;

		public void addCostObject(CostObject costObject) {
			getCostObjects().add(costObject);
		}

		public List<CostObject> getCostObjects() {
			if(costObjects == null) {
				costObjects = new ArrayList<>();
			}
			return costObjects;
		}
	}

	@Entity
	public static class Element extends BaseCostElement {
		@OneToMany(mappedBy = "parent", fetch = FetchType.EAGER)
		private List<Element> children;

		@ManyToOne
		private Element parent;

		private String type;

		public Element(String type, Element parent) {
			this.parent = parent;
			this.type = type;
		}

		public List<CostObject> getAllCostObjects() {
			ArrayList<CostObject> result = new ArrayList<>(this.getCostObjects());
			if(this.getChildren() != null) {
				this.getChildren().parallelStream().forEach(planElement -> result.addAll(planElement.getAllCostObjects()));
			}
			return result;
		}

		public long getAmountTotal() {
			ArrayList<CostObject> result = new ArrayList<>(this.getCostObjects());
			if(this.getChildren() != null) {
				this.getChildren().parallelStream().forEach(planElement -> result.addAll(planElement.getAllCostObjects()));
			}
			return result.stream().mapToLong(CostObject::getAmount).sum();
		}

		public List<Element> getChildren() {
			if(children == null) {
				this.children = new ArrayList<>();
			}
			return this.children;
		}
	}

}